/*
 * Sweep, a sound wave editor.
 *
 * Copyright (C) 2000 Conrad Parker
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <math.h>
#include <sys/ioctl.h>
#include <pthread.h>

#include <sweep/sweep_types.h>
#include <sweep/sweep_sample.h>

#include "play.h"
#include "driver.h"

#define PLAYBACK_SCALE (32768 / SW_AUDIO_T_MAX)

static sw_handle * handle = NULL;

static int playoffset = -1;
static sw_sample * prev_sample = NULL;
static pthread_t player_thread = -1;

/*
 * update_playmarker ()
 *
 * Update the position of the playback marker line for the sample
 * being played.
 *
 * Called within the main sweep interface thread.
 * gtk_idle will keep calling this function as long as this sample is
 * playing, unless otherwise stopped.
 */
static gint
update_playmarker (gpointer data)
{
  sw_sample * s = (sw_sample *)data;

  sample_set_playmarker (s, playoffset);

  if (s->play_mode == SWEEP_TRANSPORT_STOP) {
    s->playmarker_tag = 0;

    /* As this may have been stopped by the player thread, refresh the
     * interface. */
    sample_set_playmode (s, SWEEP_TRANSPORT_STOP);
    return FALSE;
  } else {
    return TRUE;
  }
}

static void
start_playmarker (sw_sample * s)
{
  if (s->playmarker_tag > 0) {
    gtk_timeout_remove (s->playmarker_tag);
  }

  s->playmarker_tag =
    gtk_timeout_add ((guint32)100,
		     (GtkFunction)update_playmarker,
		     (gpointer)s);
}

static struct timeval tv_instant = {0, 0};

static void
play_view(sw_view * view, sw_framecount_t start, sw_framecount_t end,
	  gfloat relpitch)
{
  sw_sample * s = view->sample;
  fd_set fds;
  ssize_t n = 0;
  sw_audio_t * d;
  gint16 pbuf[PBUF_SIZE];
  gint sbytes, channels;
  gint play_delta;
  gdouble po = 0.0, p, endf;
  gint i=0, si=0;

  d = (sw_audio_t *)s->sounddata->data;

  sbytes = 2;
 
  channels = s->sounddata->format->channels;

  playoffset = start;
  po = (gdouble)(start);
  endf = (gdouble)(end);

  while ((po <= endf) && sample_bank_contains (s) &&
	 s->play_mode != SWEEP_TRANSPORT_STOP) {
    FD_ZERO (&fds);
    FD_SET (handle->driver_fd, &fds);
    
    if (select (handle->driver_fd+1, NULL, &fds, NULL, &tv_instant) == 0);

    memset (pbuf, 0, sizeof (pbuf));

    switch (channels) {
    case 1:
      switch (handle->driver_channels) {
      case 1:
	for (i = 0; i < PBUF_SIZE; i++) {
	  si = (int)floor(po);
	  p = po - (gdouble)si;
	  ((gint16 *)pbuf)[i] =
	    (gint16)(PLAYBACK_SCALE * view->gain *
		     (d[si] * p + d[si+channels] * (1 - p)));
	  po += view->rate * relpitch;
	  if (po > endf) break;
	}
	break;
      case 2:
	for (i = 0; i < PBUF_SIZE; i+=2) {
	  si = (int)floor(po);
	  p = po - (gdouble)si;
	  ((gint16 *)pbuf)[i] =
	    (gint16)(PLAYBACK_SCALE * view->gain *
		     (d[si] * p + d[si+channels] * (1 - p)));
	  ((gint16 *)pbuf)[i+1] = ((gint16 *)pbuf)[i];
	  po += view->rate * relpitch;
	  if (po > endf) break;
	}
	break;
      default:
	break;
      }

      break;
    case 2:
      switch (handle->driver_channels) {
      case 1:
	/* Mix stereo down to mono */
	for (i = 0; i < PBUF_SIZE; i++) {
	  si = (int)floor(po);
	  p = po - (gdouble)si;
	  si *= 2;
	  ((gint16 *)pbuf)[i] =
	    (gint16)(PLAYBACK_SCALE * view->gain * 0.5 *
		     ((d[si] + d[si+1]) * p +
		      (d[si+channels] + d[si+1+channels]) * (1 - p)));
	  po += view->rate * relpitch;
	  if (po > endf) break;
	}
	break;
      case 2:
	for (i = 0; i < PBUF_SIZE; i++) {
	  si = (int)floor(po);
	  p = po - (gdouble)si;
	  si *= 2;
	  ((gint16 *)pbuf)[i] =
	    (gint16)(PLAYBACK_SCALE * view->gain *
		     (d[si] * p + d[si+channels] * (1 - p)));
	  i++; si++;
	  ((gint16 *)pbuf)[i] =
	    (gint16)(PLAYBACK_SCALE * view->gain *
		   (d[si] * p + d[si+channels] * (1 - p)));
	  po += view->rate * relpitch;
	  if (po > endf) break;
	}
	break;
      default:
	break;
      }
      
      break;
    }

    /* Only write if still playing */
    if (s->play_mode != SWEEP_TRANSPORT_STOP) {
      n = device_write (handle, pbuf, sizeof(pbuf));
    }

    g_mutex_lock (s->play_mutex);

    play_delta = (int)(n * view->rate * relpitch /
		       (sbytes * handle->driver_channels));

    if (playoffset != s->play_offset) {
      playoffset = s->play_offset;
      po = (gdouble) playoffset;
    } else {
      playoffset += play_delta;
      s->play_offset = playoffset;
    }

    g_mutex_unlock (s->play_mutex);
  }
}

static void
pva (sw_view * view)
{
  sw_sample * s = view->sample;

  device_setup (handle, s);

  do {

    if (s->play_offset >= s->sounddata->nr_frames) {
      g_mutex_lock (s->play_mutex);
      s->play_offset = 0;
      g_mutex_unlock (s->play_mutex);
    }

    play_view (view, s->play_offset, s->sounddata->nr_frames, 1.0);
    
  } while (s->play_mode == SWEEP_TRANSPORT_PLAY && s->play_looping);

  if (!s->play_looping) {
    device_drain (handle);

    g_mutex_lock (s->play_mutex);
    s->play_mode = SWEEP_TRANSPORT_STOP;
    g_mutex_unlock (s->play_mutex);
  }

  device_reset (handle);
  device_close (handle);
}

void
play_view_all (sw_view * view)
{
  sw_sample * s = view->sample;

  stop_playback (prev_sample);

  sample_set_playmode (s, SWEEP_TRANSPORT_PLAY);

  prev_sample = s;

  if ((handle = device_open()) != NULL) {
#ifdef DEBUG
    if (player_thread != -1) {
      g_print ("Warning: runaway player thread!\n");
    }
#endif
    pthread_create (&player_thread, NULL, (void *) (*pva), view);
    start_playmarker (s);
  } else {
    g_mutex_lock (s->play_mutex);
    s->play_mode = SWEEP_TRANSPORT_STOP;
    g_mutex_unlock (s->play_mutex);
  }
}

static void
pvs (sw_view * view)
{
  sw_sample * s = view->sample;
  GList * gl, * gl_next;
  sw_sel * sel;
  sw_framecount_t start, end;

  device_setup (handle, s);

  do {

    g_mutex_lock (s->sounddata->sels_mutex);
    for (gl = s->sounddata->sels; gl->next; gl = gl->next);

    sel = (sw_sel *)gl->data;
    end = sel->sel_end;
    g_mutex_unlock (s->sounddata->sels_mutex);

    if (s->play_offset >= end) {
      g_mutex_lock (s->play_mutex);
      s->play_offset = 0;
      g_mutex_unlock (s->play_mutex);
    }

    for (gl = s->sounddata->sels; gl; ) {
      
      if (s->play_mode != SWEEP_TRANSPORT_PLAY_SEL) break;

      /* Hold sels_mutex for as little time as possible */
      g_mutex_lock (s->sounddata->sels_mutex);
      
      /* if gl is no longer in sels, break out */
      if (g_list_position (s->sounddata->sels, gl) == -1) {
	g_mutex_unlock (s->sounddata->sels_mutex);
	break;
      }
      
      sel = (sw_sel *)gl->data;
      start = sel->sel_start;
      end = sel->sel_end;
      gl_next = gl->next;
      
      g_mutex_unlock (s->sounddata->sels_mutex);

      if (end > s->play_offset) {
	if (start < s->play_offset) {
	  start = s->play_offset;
	} else {
	  g_mutex_lock (s->play_mutex);
	  s->play_offset = start;
	  g_mutex_unlock (s->play_mutex);
	}

	play_view (view, start, end, 1.0);
      }
      
      gl = gl_next;
    }
  } while (s->play_mode == SWEEP_TRANSPORT_PLAY_SEL && s->play_looping &&
	   s->sounddata->sels != NULL);

  if (s->play_mode == SWEEP_TRANSPORT_PLAY_SEL) {
    device_drain (handle);

    g_mutex_lock (s->play_mutex);
    s->play_mode = SWEEP_TRANSPORT_STOP;
    g_mutex_unlock (s->play_mutex);
  }

  if (playoffset >= s->sounddata->nr_frames)
    sample_set_playmarker (s, 0);

  device_reset (handle);
  device_close (handle);
}

void
play_view_sel (sw_view * view)
{
  sw_sample * s = view->sample;

  if (s->sounddata->sels == NULL) {
    sample_set_playmode (s, SWEEP_TRANSPORT_STOP);
    return;
  }

  stop_playback (prev_sample);

  sample_set_playmode (s, SWEEP_TRANSPORT_PLAY_SEL);

  prev_sample = s;

  if ((handle = device_open()) != NULL) {
#ifdef DEBUG
    if (player_thread != -1) {
      g_print ("Warning: runaway player thread\n");
    }
#endif
    pthread_create (&player_thread, NULL, (void *) (*pvs), view);
    start_playmarker (s);
  } else {
    g_mutex_lock (s->play_mutex);
    s->play_mode = SWEEP_TRANSPORT_STOP;
    g_mutex_unlock (s->play_mutex);
  }
}

typedef struct _pvap_data pvap_data;
struct _pvap_data {
  sw_view * view;
  gfloat pitch;
};

static void
pvap (pvap_data * p)
{
  sw_sample * s = p->view->sample;

  device_setup (handle, s);

  g_mutex_lock (s->play_mutex);
  s->play_offset = 0;
  g_mutex_unlock (s->play_mutex);

  do {

    if (s->play_offset >= s->sounddata->nr_frames) {
      g_mutex_lock (s->play_mutex);
      s->play_offset = 0;
      g_mutex_unlock (s->play_mutex);
    }

    play_view (p->view, s->play_offset, s->sounddata->nr_frames, p->pitch);
    
  } while (s->play_mode == SWEEP_TRANSPORT_PLAY && s->play_looping);

  if (!s->play_looping) {
    device_drain (handle);

    g_mutex_lock (s->play_mutex);
    s->play_mode = SWEEP_TRANSPORT_STOP;
    g_mutex_unlock (s->play_mutex);
  }

  device_reset (handle);
  device_close (handle);

  g_free (p);
}

void
play_view_all_pitch (sw_view * view, gfloat pitch)
{
  sw_sample * s = view->sample;
  pvap_data * p;

  p = g_malloc (sizeof (*p));
  p->view = view;
  p->pitch = pitch;

  stop_playback (s);

  sample_set_playmode (s, SWEEP_TRANSPORT_PLAY);

  prev_sample = s;

  if ((handle = device_open()) != NULL) {
#ifdef DEBUG
    if (player_thread != -1) {
      g_print ("Warning: runaway player thread\n");
    }
#endif
    pthread_create (&player_thread, NULL, (void *) (*pvap), p);
    start_playmarker (s);
  } else {
    g_mutex_lock (s->play_mutex);
    s->play_mode = SWEEP_TRANSPORT_STOP;
    g_mutex_unlock (s->play_mutex);
  }
}

void
stop_all_playback (void)
{
  if (prev_sample != NULL)
    sample_set_playmode (prev_sample, SWEEP_TRANSPORT_STOP);

  device_flush (handle);

  pthread_join (player_thread, NULL);
  player_thread = -1;
}

void
stop_playback (sw_sample * s)
{
  if (s == NULL) return;

  if (s->play_mode != SWEEP_TRANSPORT_STOP) {
    sample_set_playmode (s, SWEEP_TRANSPORT_STOP);
  }

  device_flush (handle);

  pthread_join (player_thread, NULL);
  player_thread = -1;
}
